import typing

class Object(dict):
    def __init__(self, data: dict = {}):
        super().__init__(data)
        try:
            self.__recurse_to_self()
        except RecursionError:
            pass
    def __getattr__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError as e:
            raise AttributeError(e)
    def __setattr__(self, name: str, value: typing.Any) -> None:
        return super().__setitem__(name, Object(value) if type(value) == dict else value)
    def __delattr__(self, name: str):
        return super().pop(name)
    def __setitem__(self, k, value) -> None:
        return super().__setitem__(k, Object(value) if type(value) == dict else value)
    def __iter__(self):
        return super().items()
    def map(self, func: typing.Callable):
        _copy_data = super().copy()
        for key, val in super().items():
            _copy_data.__setitem__(key, func((key, val)))
        return _copy_data
    def map_key(self, func: typing.Callable):
        _copy_data = super().copy()
        for key, val in super().items():
            _copy_data.__setitem__(func((key, val)), _copy_data.pop(key))
        return _copy_data
    def map_items(self, func: typing.Callable):
        _copy_data = {}
        for key, val in super().items():
            item = func((key, val))
            _copy_data.__setitem__(item[0], item[1])
        return _copy_data
    def __recurse_to_self(self):
        for key, value in super().items():
            if type(value) == dict:
                super().__setitem__(key, Object(value))


import typing

class ParameterConflict(Exception):
    def __init__(self, message: str, *, paramters: typing.List[str]):
        super().__init__(message)
        self.message = message
        self.parameter = paramters
    def __repr__(self) -> str:
        return f"{self.message}"
    def __str__(self) -> str:
        return self.__repr__()


import typing, asyncio, inspect
from typing import overload

from .utils import to_event, enforce_type
from .models import Object
from .collector import EventCollector
from .exceptions import EventNotFound

class BaseSocket:
    def __init__(self):
        self.listeners: Object = Object({
            'message':[],
            'connect':[],
            'ready':[],
            'close':[],
            'disconnect': [],
            'message_collector': [],
            'connect_collector': [],
            'disconnect_collector':[],
            'close_collector': [],
        })
        self.events = ['message', 'connect', 'ready', 'close', 'disconnect']
        self.loop = asyncio.get_event_loop()
    @overload
    def on(self, event: str):
        ...
    @enforce_type
    def on(self, event: to_event):
        if event not in self.events:
            raise EventNotFound(f"There is no registerable event with the name \"{event}\".")
        def decorator(coro: typing.Coroutine):
            if not inspect.iscoroutinefunction(coro):
                raise TypeError(f"Expected a coroutine function found {type(coro).__name__}.")
            if not self.listeners[event]: 
                self.listeners[event] = [coro]
            else: 
                self.listeners[event].append(coro)
        return decorator
    def event(self, coro: typing.Coroutine):
        if not inspect.iscoroutinefunction(coro):
            raise TypeError(f"Expected a coroutine function found {type(coro).__name__}.")
        if coro.__name__.startswith("on_"): self.listeners[coro.__name__[3:]].append(coro)
    @overload
    def wait_for(self, event: str, *, 
                       check: typing.Optional[typing.Callable] = None, 
                       timeout: typing.Optional[float] = None):
        ...
    @enforce_type
    def wait_for(self, event: to_event, *, check = None, timeout: float = None):
        if event not in self.events: raise EventNotFound(f"There is no registerable event with the name \"{event}\".")
        future = self.loop.create_future()
        if check in [True, None, False]:
            check = lambda *args: True if check in [True, None] else False
        self.listeners[f"{event}_collector"].append((future, check))
        return asyncio.wait_for(future, timeout=timeout)
    @classmethod
    def Collector(websocket, time: float):
        return EventCollector(websocket=websocket, time=time)
    @overload
    def get_listeners(self, event: typing.Optional[str] = None) -> typing.Union[Object, typing.List[typing.Coroutine]]:
        ...
    def get_listeners(self, event: str = None):
        event = to_event(event)
        if event:
            if event not in self.events: raise EventNotFound(f"There is no registerable event with the name \"{event}\".")
            return self.listeners[event]
        else:
            return Object({event_name:listeners for event_name, listeners in self.listeners.items() if not event_name.endswith("_collector")})
    def set_listener(self, event: str, coro: typing.Coroutine):
        event = to_event(event)
        if event not in self.events: raise EventNotFound(f"There is no registerable event with the name \"{event}\".")
        if not self.listeners[event]: 
            self.listeners[event] = [coro]
        else: 
            self.listeners[event].append(coro)
        return coro


from datetime import datetime
import websockets, typing, asyncio, json
from json.decoder import JSONDecodeError
from websockets import ConnectionClosedError

from .base import BaseSocket
from .models import Message, Object
from .wsprotocols import WSCProtocol
from .collector import EventCollector


class ClientSocket(BaseSocket):
    def __init__(self):
        super().__init__()
        self.listeners["message"].append(self.on_message)
        self.listeners["connect"].append(self.on_connect)
        self.listeners["disconnect"].append(self.on_disconnect)
        self.listeners.close.append(self.on_close)
        self.connection = None
        self.disconnection = None

    def connect(self, uri: str, **kwargs):
        kwargs.pop("create_protocol", None)
        self.loop.run_until_complete(self.__main(uri, **kwargs))
        self.loop.run_forever()

    async def on_message(self, message):
        pass

    async def __message_consumer(self):
        try:
            async for message in self.connection:
                try:
                    data = json.loads(message)
                except JSONDecodeError:
                    data = message
                message_cls: Message = Message(
                    data=data, websocket=self.connection, created_at=datetime.utcnow()
                )
                async with asyncio.TaskGroup() as tg:
                    for coro in self.listeners["message"]:
                        tg.create_task(coro(message_cls))
                    for futures in self.listeners.message_collector:
                        tg.create_task(
                            self.__collector_verifier(futures, "message", message_cls)
                        )
        except ConnectionClosedError as e:
            self.disconnection = Object(
                {"code": e.code, "reason": e.reason, "disconnected": True}
            )
            self.loop.create_task(
                asyncio.wait(
                    [coro(e.code, e.reason) for coro in self.listeners.disconnect]
                    + [
                        self.__collector_verifier(
                            futures, "disconnect", e.code, e.reason
                        )
                        for futures in self.listeners.disconnect_collector
                    ]
                )
            )
            return e

    async def __on_connect(self):
        async with asyncio.TaskGroup() as tg:
            for coro in self.listeners["connect"]:
                tg.create_task(coro())

    async def on_connect(self):
        pass

    async def on_disconnect(self, code, reason):
        pass

    async def on_close(self, code, reason):
        pass

    async def __collector_verifier(self, futures, event, *event_data):
        if futures[1](*event_data):
            try:
                futures[0].set_result(
                    event_data[0] if len(event_data) == 1 else event_data
                )
                self.listeners[f"{event}_collector"].remove(futures)
            except asyncio.exceptions.InvalidStateError:
                try:
                    self.listeners[f"{event}_collector"].remove(futures)
                except ValueError:
                    pass
            except ValueError:
                pass

    def collector(self, time: float):
        return EventCollector(websocket=self, time=time)

    async def __main(self, uri, **kwargs):
        self.connection = await websockets.connect(
            uri, create_protocol=WSCProtocol, **kwargs
        )
        async with asyncio.TaskGroup() as tg:
            tg.create_task(self.__on_connect())
        done = await asyncio.wait_for(self.__message_consumer(), timeout=None)
        if ConnectionClosedError == done:
            return
        self.disconnection = Object(
            {
                "code": self.connection.close_code,
                "reason": self.connection.close_reason,
                "disconnected": True,
            }
        )
        await asyncio.wait(
            [
                coro(self.connection.close_code, self.connection.close_reason)
                for coro in self.listeners.close
            ]
            + [
                self.__collector_verifier(
                    futures,
                    "close",
                    self.connection.close_code,
                    self.connection.close_reason,
                )
                for futures in self.listeners.close_collector
            ]
        )

    async def send(self, content: typing.Any = None, *, data: dict = None):
        await self.connection.send(content=content, data=data)

    async def close(self, code: int = 1000, reason: str = ""):
        await self.connection.close(code=code, reason=reason)



import typing, inspect, asyncio
from inspect import Parameter

def enforce_type(func: typing.Union[typing.Callable, typing.Coroutine]):
    def wrapper(*args, **kwargs):
        defs, args, kwargs = get_defaults(func), list(args).copy(), kwargs.copy()
        for key, val in defs.items():
            if val[1] in [Parameter.POSITIONAL_ONLY, Parameter.VAR_POSITIONAL, Parameter.POSITIONAL_OR_KEYWORD] and len(args) > val[2] and key not in kwargs:
                args.insert(val[2], val[0])
            elif val[1] in [Parameter.KEYWORD_ONLY, Parameter.VAR_KEYWORD] and key not in kwargs:
                kwargs[key] = val[0]
        annotation_vals = {list(func.__code__.co_varnames).index(key):vals for key, vals in func.__annotations__.items()}
        args = [annotation_vals[j](i) if j in list(annotation_vals.keys()) and i != None else i for i,j in zip(args, range(len(args)))]
        kwargs = {key:func.__annotations__[key](value) if key in func.__annotations__ and value != None else value for key, value in kwargs.items()}
        if inspect.iscoroutinefunction(func): return asyncio.run(func(*args, **kwargs))
        else: return func(*args, **kwargs)
    return wrapper

def get_defaults(func: typing.Union[typing.Callable, typing.Coroutine]):
    sig = inspect.signature(func)
    return {
        item[0]: [item[1].default, item[1].kind, ind] for item, ind in zip(sig.parameters.items(), range(len(sig.parameters.items()))) 
        if item[1].default is not inspect.Parameter.empty
    }


from datetime import datetime
import typing, websockets, asyncio, json
from json.decoder import JSONDecodeError
from websockets import ConnectionClosedError

from .base import BaseSocket
from .models.message import Message, Object
from .wsprotocols import WSSProtocol
from .collector import EventCollector

class ServerSocket(BaseSocket):
    def __init__(self):
        super().__init__()
        self.listeners.message.append(self.on_message)
        self.listeners.connect.append(self.on_connect)
        self.listeners.ready.append(self.on_ready)
        self.listeners.disconnect.append(self.on_disconnect)
        self.listeners.close.append(self.on_close)
        self.clients = []
        self.disconnected_clients = []
    def listen(self, addr:str, port:int, **kwargs):
        kwargs.pop('create_protocol', None)
        self.address = addr
        self.port = port
        self.server = websockets.serve(self.__main, addr, port, create_protocol = WSSProtocol, **kwargs)
        self.loop.run_until_complete(self.server)
        self.loop.run_until_complete(asyncio.wait([coro() for coro in self.listeners.ready if type(coro) != tuple]))
        self.loop.run_forever()
    async def on_message(self, message):
        pass
    async def __message_consumer(self, websocket):
        try:
            async for message in websocket:             
                try:
                    data = json.loads(message)
                except JSONDecodeError:
                    data = message
                message_cls: Message = Message(data=data, websocket=websocket, created_at=datetime.utcnow())
                self.loop.create_task(asyncio.wait(
                     [coro(message_cls) for coro in self.listeners.message]+[self.__collector_verifier(futures, 'message', message_cls) 
                     for futures in self.listeners.message_collector
                    ]))
        except ConnectionClosedError as e:
            self.clients.remove(websocket)
            self.disconnected_clients.append({websocket: Object({'code': e.code, 'reason': e.reason, 'disconnected': True})})
            self.loop.create_task(asyncio.wait([coro(websocket, e.code, e.reason) for coro in self.listeners.disconnect]+[
                     self.__collector_verifier(futures, 'disconnect', e.code, e.reason) 
                     for futures in self.listeners.disconnect_collector
                    ]))
            return e
    async def __on_connect(self, client, path):
        await asyncio.wait([coro(client, path) for coro in self.listeners.connect]+[self.__collector_verifier(futures, 'connect', client, path) 
                     for futures in self.listeners.connect_collector
                    ])
    async def on_connect(self, client, path):
        pass
    async def on_disconnect(self, client, code, reason):
        pass
    async def on_close(self, client, code, reason):
        pass
    async def on_ready(self):
        pass
    async def __collector_verifier(self, futures, event, *event_data):
        if futures[1](*event_data):
            try: 
                futures[0].set_result(event_data[0] if len(event_data) == 1 else event_data)
                self.listeners[f"{event}_collector"].remove(futures)
            except asyncio.exceptions.InvalidStateError:
                try:
                    self.listeners[f"{event}_collector"].remove(futures)
                except ValueError:
                    pass
            except ValueError:
                pass
    def collector(self, time: float):
        return EventCollector(websocket=self, time=time)
    async def __main(self, websocket, path):
        self.clients.append(websocket)
        self.loop.create_task(self.__on_connect(websocket, path))
        done, pending = await asyncio.wait([self.__message_consumer(websocket)], return_when=asyncio.ALL_COMPLETED)
        if ConnectionClosedError in [type(ret.result()) for ret in done]: return
        self.clients.remove(websocket)
        self.disconnected_clients.append({websocket: Object({'code': websocket.close_code, 'reason': websocket.close_reason, 'disconnected': True})})
        await asyncio.wait([coro(websocket, websocket.close_code, websocket.close_reason) for coro in self.listeners.close]+[
                     self.__collector_verifier(futures, 'close', websocket, websocket.close_code, websocket.close_reason) 
                     for futures in self.listeners.close_collector
                    ])
    async def send(self, client, content: typing.Any = None, *, data: dict = None):
        await client.send(content=content, data=data)
    async def close(self, client, *, code: int = 1000, reason: str = ''):
        await client.close(code=code, reason=reason)


import typing, datetime, asyncio

class EventCollector:
    def __init__(self, websocket, time: float) -> None:
        self.__websocket = websocket
        self.__time = time
    async def collect(self, event: str, check: typing.Callable = None) -> typing.List[typing.Any]:
        collected = []
        start_time = datetime.datetime.utcnow()
        while (datetime.datetime.utcnow() - start_time).total_seconds() < self.__time:
            try:
                event_data = await self.__websocket.wait_for(event, 
                                                             check=check, 
                                                             timeout=self.__time - (datetime.datetime.utcnow()-start_time).total_seconds())                
                collected.append(event_data)
            except asyncio.TimeoutError:
                pass
        return collected


import datetime

def to_event(eve: str):
    return eve.strip().lower()

def datetime_to_dict(dt: datetime.datetime):
    return {
		'year':dt.year,
		'month':dt.month,
		'day':dt.day,
		'hour':dt.hour,
		'minute':dt.minute,
		'second':dt.second,
		'millisecond':dt.microsecond
	}


from typing import Literal, NamedTuple

from .client import ClientSocket
from .server import ServerSocket
from .models import Message, Object
from .wsprotocols import WSSProtocol, WSCProtocol
from .exceptions import ParameterConflict, EventNotFound
from .collector import EventCollector

class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]
    serial: int

__version__: str = "2.1.3"
__author__: str = "Cybertron"
version_info: VersionInfo = VersionInfo(major=2, minor=1, micro=3, releaselevel='final', serial=0)
__copyright__: str = "Copyright (c) 2021 - present Cybertron"


import datetime, typing

from .data import Object
from ..wsprotocols import WSSProtocol, WSCProtocol

class Message:
    def __init__(self, data: typing.Any = {}, 
                       websocket: typing.Union[WSSProtocol, WSCProtocol] = None, 
                       created_at: datetime.datetime = None):
        self.data: typing.Any = Object(data) if type(data) == dict else data
        self.author: typing.Union[WSSProtocol, WSCProtocol, None] = websocket
        self.created_at: datetime.datetime = created_at
    def __repr__(self) -> str:
        return f"<Message {'message' if type(self.data) != Object else 'data'}={self.data} author={self.author} created_at={self.created_at}>"
    def __str__(self) -> str:
        return str(self.data)


import websockets, typing, json
from ..exceptions import ParameterConflict

class WSCProtocol(websockets.WebSocketClientProtocol):
    def __init__(self, *args, **kwargs):
        """
            A subclass of WebSocketClientProtocol to provide extended send and receive functionalities. 
        """
        super().__init__(*args, **kwargs)
    async def send(self, content: typing.Union[str, bytes, bytearray, typing.List[typing.Any]] = None, *, data: dict = None):
        if not data and not content:
            raise TypeError("Missing either both of \"data\" or \"content\" parameter. Any one is required.")
        if data and content:
            raise ParameterConflict("Got values for both data and content. Expected anyone.", paramters=['data', 'content'])
        if not data:
            await super().send(content)
            return
        try:
            await super().send(json.dumps((dict(data))))
        except ValueError:
            raise ValueError(f"\"data\" parameter expects a dictionary or a dictionary convertable object. Got {type(data).__name__}")
        except TypeError as e:
            raise TypeError(". ".join(list(e.args)))
    def __repr__(self) -> str:
        return f"<WSClientProtocol remote_address={self.remote_address} local_address={self.local_address}>"
    def __str__(self) -> str:
        return self.__repr__()
